@@ -16,13 +16,17 @@
  *
  */
 
-#include "common.h"
+#include "epos.h"
 
-#define DIPH_BUFF_SIZE  1000 //unimportant
+#ifdef HAVE_WINDOWS_H
+	#include <windows.h>
+#endif
 
-#define OPCODEstr "subst:regex:postp:prep:segments:prosody:contour:progress:regress:insert:syll:smooth:raise:debug:if:inside:with:{:}:[:]:<:>:nothing:error:"
-enum OPCODE {OP_SUBST, OP_REGEX, OP_POSTP, OP_PREP, OP_DIPH, OP_PROSODY, OP_CONTOUR, OP_PROGRESS, OP_REGRESS, 
-	OP_INSERT, OP_SYLL, OP_SMOOTH, OP_RAISE, OP_DEBUG, OP_IF, OP_INSIDE, OP_WITH,
+#define SEG_BUFF_SIZE  1000 //unimportant
+
+#define OPCODEstr "nnet:subst:regex:postp:prep:segments:absolutize:prosody:contour:progress:regress:insert:syll:analyze:smooth:raise:debug:if:inside:near:with:{:}:[:]:<:>:nothing:error:"
+enum OPCODE {OP_NNET, OP_SUBST, OP_REGEX, OP_POSTP, OP_PREP, OP_SEG, OP_ABSOL, OP_PROSODY, OP_CONTOUR, OP_PROGRESS, OP_REGRESS, 
+	OP_INSERT, OP_SYLL, OP_ANALYZE, OP_SMOOTH, OP_RAISE, OP_DEBUG, OP_IF, OP_INSIDE, OP_NEAR, OP_WITH,
 	OP_BEGIN, OP_END, OP_CHOICE, OP_CHOICEND, OP_SWITCH, OP_SWEND, OP_NOTHING, OP_ERROR};
 		/* OP_BEGIN, OP_END and other OP's without parameters should come last
 		   OP_ERROR would abort the compilation (never used)			*/
@@ -55,10 +59,12 @@ extern char * _rules_buff;
 #define SPACE		' '	// to delimit lists in literal hash tables
 #define TILDE		'~'
 
-#define UNDIPHONE "~"			//OP_DIPH only, means "delete segments if any"
+#define UNSEGMENT "~"			//OP_SEG only, means "delete segments if any"
 
 class rule
 {
+   private:
+	virtual void apply(unit *root) = 0;
    protected:
 	const char *debug_tag();
    public:
@@ -74,8 +80,9 @@ class rule
 	virtual void check_child(rule *r);
 	virtual void check_children();
 	virtual void verify() {};
-	virtual void apply(unit *root) = 0;
 	virtual void debug();
+
+	void cook(unit *root);
 #ifdef DEBUGGING
 	char *dbg_tag;
 #endif
@@ -85,7 +92,10 @@ class hashing_rule: public rule
 {
    protected:
 	bool allow_id;
+	bool use_fastmatch;
+	char negated;
 	hash *dict;
+	regex_t *fastmatch;
    public:
 		hashing_rule(char *param);
 	virtual ~hashing_rule();
@@ -103,7 +113,7 @@ rule::rule(char *param)
 
 rule::~rule()
 {
-	DEBUG(0,1,fprintf(STDDBG, "rule::~rule, raw=%s, dbg_tag=%s\n", raw, dbg_tag););
+	D_PRINT(0, "rule::~rule, raw=%s, dbg_tag=%s\n", raw, dbg_tag);
 	if (raw) free(raw);
 #ifdef DEBUGGING
 	if (dbg_tag) free(dbg_tag);
@@ -113,17 +123,41 @@ rule::~rule()
 void
 rule::set_level(UNIT scp, UNIT trg)
 {
-	scope =  scp == U_DEFAULT ? cfg->default_scope  : scp;
-	target = trg == U_DEFAULT ? cfg->default_target : trg;
-	if (scope < target) shriek (811, fmt("Scope must not be more narrow than target%s", debug_tag()));
+	scope =  scp == U_DEFAULT ? scfg->default_scope  : scp;
+	target = trg == U_DEFAULT ? scfg->default_target : trg;
+	if (scope < target) shriek (811, "%s Scope must not be more narrow than target", debug_tag());
 			// was scope <= target -  consider returning back
 }
 
+void
+rule::cook(unit *root)
+{
+	bool tmpscope;
+	unit *u;
+#ifdef DEBUGGING
+		if (scfg->debug) current_debug_tag = dbg_tag;
+		if (scfg->show_rule) fprintf(STDDBG, "[%s] %s %s\n",
+			dbg_tag,
+			enum2str(code(), OPCODEstr), raw);
+#endif
+	for (u = root->LeftMost(scope); u != &EMPTY;) {
+		unit *tmp_next = u->Next(scope);
+		tmpscope = u->scope;
+		u->scope = true;
+		apply(u);
+		u->scope = tmpscope;
+		u = tmp_next;
+	}
+#ifdef DEBUGGING
+		current_debug_tag = NULL;
+#endif
+}
+
 void
 rule::set_dbg_tag(text *file)
 {
 #ifdef DEBUGGING
-	DEBUG(0,1,fprintf(STDDBG,"Debugging tag %s\n",file->current_file));
+//	D_PRINT(0, "Debugging tag %s\n", file->current_file);
 	sprintf(scratch, "%s:%d", file->current_file, file->current_line);
 	dbg_tag=strdup(scratch);
 #else
@@ -138,7 +172,7 @@ rule::check_child(rule *r)
 		if (r->scope != U_INHERIT) {
 			printf("this %d child %d\n",scope, r->scope);
 			debug();
-			shriek(811, fmt("Scope of rule exceeds scope of its block%s", debug_tag()));
+			shriek(811, "%s Scope of rule exceeds scope of its block", debug_tag());
 		}
 		r->scope = this->scope;
 	}
@@ -152,7 +186,7 @@ rule::check_children()
 const char *
 rule::debug_tag()
 {
-	char *wholetag=(char *)FOREVER(xmalloc(cfg->max_line));
+	static char *wholetag=(char *)FOREVER(xmalloc(scfg->max_line_len));
 
 #ifdef DEBUGGING
 	char *tmp;
@@ -162,14 +196,14 @@ rule::debug_tag()
 			sprintf(wholetag, "in unnumbered place %s", dbg_tag);
 		} else {
 			*tmp++=0;
-			sprintf(wholetag, " in file %s line %s", dbg_tag, tmp);
+			sprintf(wholetag, "%s:%s", dbg_tag, tmp);
 			tmp[-1]=':';
 		}
 	} else
 #endif
 	{
 		if (!global_current_file) return ". Debug tag is corrupted as well.";
-		sprintf(wholetag, " in file %s line %d", global_current_file, global_current_line);
+		sprintf(wholetag, "%s:%d", global_current_file, global_current_line);
 	}
 	return wholetag;
 }
@@ -215,51 +249,144 @@ hash *literal_hash(char *s)
 	}
 }
 
+#if 0
+static char esc(char x)
+{
+	return esctab->xlat(x);
+}
+#endif
+
 hashing_rule::hashing_rule(char *param) : rule(NULL)
 {
-	if (*param == DQUOT) raw = strdup(param);
-	else raw = compose_pathname(param, this_lang->hash_dir, cfg->lang_base_dir);
+	negated = 0;
+	if (*param == '!') param++, negated = 1;
+	raw = strdup(param);
+//	if (*param == DQUOT) raw = strdup(param);
+//	else raw = compose_pathname(param, this_lang->hash_dir, scfg->lang_base_dir);
 	dict = NULL;
+	fastmatch = NULL;
 	allow_id = false;
+	use_fastmatch = false;
 }
 
 
 hashing_rule::~hashing_rule()
 {
 	if (dict) delete dict;
+#ifdef WANT_REGEX
+	if (fastmatch) {
+		regfree(fastmatch);
+		free(fastmatch);
+	}
+#endif
 }
 
 void
 hashing_rule::verify()
 {
+	if (negated > 1)
+		shriek(811, "%s Unexpected negated dictionary", debug_tag());
 	if (cfg->paranoid) {
 		load_hash();
-		if (cfg->lowmemory) {
-			DEBUG(2,2,fprintf(STDDBG,"Hash table caching is disabled.\n");) //hashtabscache[rulist[i].param]->debug();
+
+		if (scfg->memory_low) {
+			D_PRINT(2, "Hash table caching is disabled.\n"); //hashtabscache[rulist[i].param]->debug();
 			delete dict;
 			dict = NULL;
 		}
 	}
 }
 
+#ifdef WANT_REGEX
+
+#define REGEX_SPECIALS	".[*\\"	// contains no ^ nor $ nor { nor ( ) + = |
+
+static bool begins_with_caret(char *key) { return key[0] == '^'; }
+static bool ends_with_dollar(char *key) { return *key && key[strlen(key) - 1] == '$'; }
+static int special_chars(char *key)
+{
+	int r = 0;
+	for (char *s = key; *s; s++) if (strchr(REGEX_SPECIALS, *s)) r++;
+	return r;
+}
+
+void strcpy_escapeful(char *where, const char *from)
+{
+	do {
+		if (*from && strchr(REGEX_SPECIALS, *from)) *where++ = '\\';
+	} while ((*where++ = *from++));
+}
+
+static void compute_size(char *key, char *value, void *total)
+{
+	*(int *)total += strlen(key) + 2 + special_chars(key)
+		+ 2 * begins_with_caret(key) + 2 * ends_with_dollar(key);
+}
+
+static void append_to_regex(char *key, char *value, void *str)
+{
+	char *s = (char *)str;
+	int l = strlen(s);
+
+	if (begins_with_caret(key)) {
+		strcpy(s + l, "^\\");
+		l += 2;
+	}
+	strcpy_escapeful(s + l, key);
+	l += strlen(key) + special_chars(key);
+
+	if (ends_with_dollar(key)) {
+		strcpy(s + l - 1, "\\$$");
+		l += 2;
+	}
+	strcpy(s + l, "\\|");
+}
+
+static regex_t *regex_from_hash(hash *h)
+{
+	int total = 3;	
+	h->forall(compute_size, &total);
+	char *b = (char *)xmalloc(total);
+	*b = 0;
+	h->forall(append_to_regex, b);
+	if (*b) b[strlen(b) - 2] = 0;
+	D_PRINT(1, "Fastmatch regex: %s\n", b);
+	regex_t *r = (regex_t *)xmalloc(sizeof(regex_t));
+	if (regcomp(r, b, 0)) {
+		D_PRINT(3, "Failed to create a fastmatch from %.20s..., code=%d\n", b, regcomp(r, b, 0));
+		return 0;
+	}
+	free(b);
+	return r;
+}
+
+#endif
+
+
 void
 hashing_rule::load_hash()
 {
 	if (dict) shriek(862, "unwanted load_hash");
 
+	dict = NULL;
 	if (*raw != DQUOT) {
-		dict = new hash(raw, cfg->hash_full, 0, 200, 3,
-			(char *) allow_id, false, "dictionary %s not found", esctab);
+		dict = new hash(raw, this_lang->hash_dir, scfg->lang_base_dir, "dictionary",
+			scfg->hashes_full, 0, 200, 5,
+			(char *) allow_id, false);
 	} else dict = literal_hash(raw);
-	if (!dict) shriek(463, fmt("Unterminated argument%s", debug_tag()));	// or out of memory
+#ifdef WANT_REGEX
+	if (use_fastmatch)
+		fastmatch = regex_from_hash(dict);
+#endif
+	if (!dict) shriek(463, "%s Unterminated argument", debug_tag());	// or out of memory
 }
 
 void 
 rule::debug()
 {
 	fprintf(STDDBG," rule: '%s' ",enum2str(code(),OPCODEstr));
-	fprintf(STDDBG,"par %s scope '%s' ", raw, enum2str(scope, cfg->unit_levels));
-	fprintf(STDDBG,"target '%s'\n", enum2str(target, cfg->unit_levels));
+	fprintf(STDDBG,"par %s scope '%s' ", raw, enum2str(scope, scfg->unit_levels));
+	fprintf(STDDBG,"target '%s'\n", enum2str(target, scfg->unit_levels));
 }
 
 
@@ -300,25 +427,36 @@ class r_postp: public r_subst
 r_subst::r_subst(char *param) : hashing_rule(param)
 {
 	method = M_SUBSTR;
+	negated *= 2;
+	use_fastmatch = true;
 }
 
 r_prep::r_prep(char *param) : r_subst(param)
 {
 	method = M_RIGHT;
+	if (negated) {
+		method = (SUBST_METHOD)(method | M_NEGATED);
+		negated = 1;
+	}
 	allow_id = true;
+	use_fastmatch = false;
 }
 
 r_postp::r_postp(char *param) : r_subst(param)
 {
 	method = M_LEFT;
+	if (negated) {
+		method = (SUBST_METHOD)(method | M_NEGATED);
+		negated = 1;
+	}
 	allow_id = true;
+	use_fastmatch = false;
 }
 
 void
 r_subst::set_level(UNIT scp, UNIT trg)
 {
 	rule::set_level(scp, trg);
-//	if (target != U_PHONE) shriek(fmt("Cannot substitute for non-phones %s",debug_tag()));
 }
 
 /************************************************
@@ -332,10 +470,10 @@ r_subst::apply(unit *root)
 
 //	if (target == U_PHONE) root->subst(dict, method);
 
-	root->relabel(dict, method, target);
+	root->relabel(dict, fastmatch, method, target);
 
-	if (cfg->lowmemory) {
-		DEBUG(2,2,fprintf(STDDBG,"Hash table caching is disabled.\n");) //hashtabscache[rulist[i].param]->debug();
+	if (scfg->memory_low) {
+		D_PRINT(2, "Hash table caching is disabled.\n"); //hashtabscache[rulist[i].param]->debug();
 		delete dict;
 		dict = NULL;
 	}
@@ -350,7 +488,7 @@ r_subst::apply(unit *root)
 
 class r_seg: public hashing_rule
 {
-	virtual OPCODE code() {return OP_DIPH;};
+	virtual OPCODE code() {return OP_SEG;};
    public:
 		r_seg(char *param);
 	virtual void apply(unit *root);
@@ -358,7 +496,7 @@ class r_seg: public hashing_rule
 
 r_seg::r_seg(char *param) : hashing_rule(param)
 {
-	if (!strcmp(param, UNDIPHONE)) {
+	if (!strcmp(param, UNSEGMENT)) {
 		free(raw);
 		raw=strdup(NULL_FILE);
 	}
@@ -373,25 +511,76 @@ r_seg::apply(unit *root)
 {
 	if (!dict) load_hash();
 	root->segs(target, dict);
-	if (cfg->lowmemory) {
-		DEBUG(2,2,fprintf(STDDBG,"Hash table caching is disabled.\n");) //hashtabscache[rulist[i].param]->debug();
+	if (scfg->memory_low) {
+		D_PRINT(2, "Hash table caching is disabled.\n"); //hashtabscache[rulist[i].param]->debug();
 		delete dict;
 		dict=NULL;
 	}
 
-	DEBUG(1,2,fprintf(STDDBG,"Diphones w%s be dumped just after the DIPHONES rule\n", cfg->imm_segs?"ill":"on't");)
-	if (cfg->imm_segs) {
-		static segment d[DIPH_BUFF_SIZE];   //every item is 16 bytes long
+
+	D_PRINT(1, "Segments w%s be dumped just after the segments rule\n", scfg->immed_segments?"ill":"on't");
+	if (scfg->immed_segments) {
+		static segment d[SEG_BUFF_SIZE];   //every item is 16 bytes long
 		
-		int i=DIPH_BUFF_SIZE;
-		for (int k=0; i==DIPH_BUFF_SIZE; k+=DIPH_BUFF_SIZE) {
-			i=root->write_segs(d,k,DIPH_BUFF_SIZE);
+		int i = SEG_BUFF_SIZE;
+		for (int k=0; i == SEG_BUFF_SIZE; k += SEG_BUFF_SIZE) {
+			i = root->write_segs(d, k, SEG_BUFF_SIZE);
 			for (int j=0;j<i;j++) 
 				fprintf(STDDBG,"segment number=%3d f=%d t=%d i=%d\n", d[j].code, d[j].f, d[j].t, d[j].e);
 		}
 	}
 }
 
+
+
+
+
+
+
+
+
+
+/************************************************
+ r_absol  The following rule class computes
+ 	  the absolute timings for target units
+ **	  using voice dependent data
+ ************************************************/
+
+
+class r_absol: public rule
+{
+	virtual OPCODE code() {return OP_ABSOL;};
+   public:
+		r_absol(char *param);
+	virtual void apply(unit *root);
+};
+
+r_absol::r_absol(char *param) : rule(param)
+{
+	if (strcmp(param, "t")) shriek(811, "%s: Impossible absolutize request", debug_tag());
+}
+
+/************************************************
+ r_absol::apply()
+ ************************************************/
+
+void
+r_absol::apply(unit *root)
+{
+	root->absol(target);
+}
+
+
+
+
+
+
+
+
+
+
+
+
 /************************************************
  r_prosody The following rule class manipulates
 	   suprasegmentalia according to subrules
@@ -409,6 +598,7 @@ class r_prosody: public hashing_rule
 
 r_prosody::r_prosody(char *param) : hashing_rule(param)
 {
+	negated *= 2;
 }
 
 /************************************************
@@ -419,10 +609,11 @@ void
 r_prosody::apply(unit *root)
 {
 	if (!dict) load_hash();
-	DEBUG(1,1,fprintf(STDDBG,"entering rules::sseg()\n");)
-	root->sseg(target, dict);
-	if (cfg->lowmemory) {
-		DEBUG(2,2,fprintf(STDDBG,"Hash table caching is disabled.\n");) //hashtabscache[rulist[i].param]->debug();
+	D_PRINT(1, "entering rules::sseg()\n");
+//	root->sseg(target, dict);
+	shriek(862, "no sseg");
+	if (scfg->memory_low) {
+		D_PRINT(2, "Hash table caching is disabled.\n"); //hashtabscache[rulist[i].param]->debug();
 		delete dict;
 		dict = NULL;
 	}
@@ -450,7 +641,7 @@ class r_contour: public rule
 r_contour::r_contour(char *param) : rule(param)
 {
 	char *p;
-	short int tmp=0, sgn=1;
+	int tmp=0, sgn=1;
 	
 	contour = (int *)xmalloc(strlen(param)*sizeof(int));
 	contour[0] = l = 0;
@@ -459,13 +650,13 @@ r_contour::r_contour(char *param) : rule(param)
 		switch (*p) {
 		case ':':  contour[l++] += tmp*sgn; tmp=0;
 				sgn=1; contour[l] = 0;  break;
-		case '*':  if (p[1] && p[1] != ':') shriek(811, fmt("A ':' should follow '*'%s", debug_tag()));
-			   if (padd_start > -1) shriek(811, fmt("Ambiguous padding%s", debug_tag()));
+		case '*':  if (p[1] && p[1] != ':') shriek(811, "%s A ':' should follow '*'", debug_tag());
+			   if (padd_start > -1) shriek(811, "%s Ambiguous padding", debug_tag());
 				padd_start = l; break;
 		case '-':
 		case '+':  contour[l]+=tmp*sgn; sgn=(*p=='+' ?+1:-1); break;
-		default:   if (*p<'0' || *p>'9') shriek(811, fmt("Expected a number, found \"%s\"%s",
-					p, debug_tag()));
+		default:   if (*p<'0' || *p>'9') shriek(811, "%s Expected a number, found \"%s\"",
+					debug_tag(), p);
 			   else tmp = tmp*10 + *p-'0';
 		}
 	}
@@ -509,7 +700,7 @@ class r_smooth: public rule
 r_smooth::r_smooth(char *param) : rule(param)
 {
 	char *p;
-	short int tmp=0, sgn=1, total=0, max;
+	int tmp=0, sgn=1, total=0, max;
 	
 	list=(int *)xmalloc(strlen(param)*sizeof(int));
 	list[0]=n=l=0;
@@ -520,14 +711,14 @@ r_smooth::r_smooth(char *param) : rule(param)
 			   sgn=1; list[l]=0;  break;
 		case '-':
 		case '+':  list[l]+=tmp*sgn; sgn=(*p=='+' ?+1:-1); break;
-		default:   if (*p<'0' || *p>'9') shriek(811, fmt("Expected a number, found \"%s\"%s",
-					p, debug_tag()));
+		default:   if (*p<'0' || *p>'9') shriek(811, "%s Expected a number, found \"%s\"",
+					debug_tag(), p);
 			   else tmp = tmp*10 + *p-'0';
 		}
 	}
 	if (tmp) list[l++]+=tmp*sgn, total+=tmp*sgn;
 	if (total!=RATIO_TOTAL)
-		shriek (811, fmt("Smooth percentages don't add up to 100%% (%d%%)%s", total, debug_tag()));
+		shriek (811, "%s Smooth percentages don't add up to 100%% (%d%%)", debug_tag(), total);
 	if (cfg->paranoid) {
 		for (tmp=max=0; tmp<l; tmp++)
 			if (list[tmp]>max) max=list[tmp];
@@ -553,8 +744,8 @@ void
 r_smooth::apply(unit *root)
 {
 	
-	DEBUG(1,1,fprintf(STDDBG,"entering rules::smooth()\n");)
-	root->project(target,0,0,0);
+	D_PRINT(1, "entering rules::smooth()\n");
+	root->project(target);
 	root->smooth(target, list, n, l, quantity);
 }
 
@@ -565,15 +756,32 @@ r_smooth::apply(unit *root)
  **	  one and in what environment
  ************************************************/
 
+static void switch_zeroes(char *s)
+{
+	int i, j;
+	for (i = 0, j = 0; s[i]; ) {
+		if (s[i] == LITERAL_ZERO) {
+			if (i && s[i - 1] == ESCAPE)
+				j--;
+			else {
+				s[j++] = ABSENT_CHARACTER;
+				i++;
+				continue;
+			}
+		}
+		s[j++] = s[i++];
+	}
+	s[j] = 0;
+}
 
 class r_regress: public rule
 {
    protected:
 	virtual OPCODE code() {return OP_REGRESS;};
 
-	bool *ltab;       //callocced by booltab()
-	bool *rtab;
-	char *fn;
+	charclass *ltab;       //callocced by booltab()
+	charclass *rtab;
+	charxlat *fn;
 	bool backwards;
 
    public:
@@ -589,6 +797,7 @@ class r_progress: public r_regress
 		r_progress(char *param);
 };
 
+
 r_regress::r_regress(char *param) : rule(param)
 {
 	char *aff;
@@ -596,31 +805,48 @@ r_regress::r_regress(char *param) : rule(param)
 	char *left;      
 	char *right;
 
-	eff=strchr(aff=strdup(param),ASSIM_DELIM1);if(!eff) shriek(811, fmt("Bad param%s", debug_tag()));
-	*eff++=0;
-	left=strchr(eff,ASSIM_DELIM2);if(!left) shriek(811, fmt("Bad param%s", debug_tag()));
-	*left++=0;
-	right=strchr(left,ASSIM_DELIM3);if(!right) shriek(811, fmt("Bad param%s", debug_tag()));
-	*right++=0;
-	fn=strchr(right,ASSIM_DELIM4);if(!fn) shriek(811, fmt("Bad param%s", debug_tag()));
-	*fn++=0;if(*fn) shriek(811, fmt("Strange appendix to param%s", debug_tag()));
-	DEBUG(0,4,fprintf(STDDBG,"Parsed assim param \"%s>%s(%s_%s)\"\n",aff,eff,left,right);)
-	fn=fntab(aff,eff);ltab=booltab(left);rtab=booltab(right);
+	eff = strchr(aff=strdup(param),ASSIM_DELIM1);
+	if (!eff) shriek(811, "%s Bad param", debug_tag());
+	*eff++ = 0;
+	switch_zeroes(aff);
+
+	left = strchr(eff,ASSIM_DELIM2);
+	if(!left) shriek(811, "%s Bad param", debug_tag());
+	*left++ = 0;
+	switch_zeroes(eff);
+	
+	right = strchr(left,ASSIM_DELIM3);
+	if (!right) shriek(811, "%s Bad param", debug_tag());
+	*right++ = 0;
+	switch_zeroes(left);
+
+
+	char *tmp = strchr(right,ASSIM_DELIM4);
+	if (!tmp) shriek(811, "%s Bad param", debug_tag());
+	*tmp++ = 0;
+	switch_zeroes(right);
+	if (*tmp) shriek(811, "%s Strange appendix to param", debug_tag());
+	
+	D_PRINT(0, "Parsed assim param \"%s>%s(%s_%s)\"\n",aff,eff,left,right);
+
+	if (strlen(aff) != strlen(eff) && strlen(eff) != 1)
+		shriek(811, "%s Bad param", debug_tag());
+	fn = new charxlat(aff,eff,true); ltab = new charclass(left); rtab = new charclass(right);
 	free(aff);
 	
-	backwards=true;
+	backwards = true;
 }
 
 r_regress::~r_regress()
 {
-	free(fn);
-	free(ltab);
-	free(rtab);
+	delete fn;
+	delete ltab;
+	delete rtab;
 }
 
 r_progress::r_progress(char *param) : r_regress(param)
 {
-	backwards=false;
+	backwards = false;
 }
 
 /********************************************************
@@ -632,8 +858,8 @@ r_regress::apply(unit *root)
 {
 	root->assim(target,backwards,fn,ltab,rtab);
 
-	if (fn[JUNCTURE] != DELETE_ME)
-		root->insert(target,backwards,fn[JUNCTURE],ltab,rtab);
+	if (fn->xlat(JUNCTURE) != JUNCTURE)
+		root->insert(target, backwards, fn->xlat(JUNCTURE), ltab, rtab);
 }
 
 /************************************************
@@ -668,9 +894,9 @@ r_syll::r_syll(char *param) : rule(param)
 	for (tmp = param; *tmp && lv; tmp++) {
 		if (*tmp==LESS_THAN) lv++;
 		else son[(unsigned char)(*tmp)]=lv;
-		DEBUG(0,1,fprintf(STDDBG,"Giving to %c sonority %d\n", *tmp, lv);)
+		D_PRINT(0, "Giving to %c sonority %d\n", *tmp, lv);
 	}
-	DEBUG(0,1,fprintf(STDDBG,"rules::parse_syll going to call syllabify()\n");)
+	D_PRINT(0, "rules::parse_syll going to call syllabify()\n");
 }
 
 r_syll::~r_syll()
@@ -678,14 +904,6 @@ r_syll::~r_syll()
 	free(son);
 }
 
-//void
-//r_syll::set_level(UNIT scp, UNIT trg)
-//{
-//	if (scp == U__DEFAULT) scp = U__SYLL;
-//	rule::set_level(scp, trg);
-//}	
-
-
 /********************************************************
  r_syll::apply
  ********************************************************/
@@ -696,6 +914,57 @@ r_syll::apply(unit *root)
 	root->syllabify(target, son);
 }
 
+/************************************************
+ r_analyze   The following rule class splits words
+ 	     into morphemes using a dictionary
+ **	     of available morphemes
+ ************************************************/
+
+
+class r_analyze: public hashing_rule
+{
+	virtual OPCODE code() { return OP_ANALYZE; };
+	int unanal_unit_penalty, unanal_part_penalty;
+   public:
+		r_analyze(char *param);
+	virtual ~r_analyze();
+	virtual void apply(unit *root);
+	virtual void verify();
+};
+
+r_analyze::r_analyze(char *param) : hashing_rule(param)
+{
+}
+
+r_analyze::~r_analyze()
+{
+}
+
+void
+r_analyze::verify()
+{
+	hashing_rule::verify();
+
+	char *uup = dict->translate("!META_unanal_unit_penalty");
+	char *upp = dict->translate("!META_unanal_part_penalty");
+	if (!uup || !upp)
+		shriek(811, "%s Must specify penalties in the dictionary", debug_tag());
+	unanal_unit_penalty = atoi(uup);
+	unanal_part_penalty = atoi(upp);
+}
+
+/********************************************************
+ r_analyze::apply
+ ********************************************************/
+
+void
+r_analyze::apply(unit *root)
+{
+	root->analyze(target, dict, unanal_unit_penalty, unanal_part_penalty);
+}
+
+
+
 /************************************************
  r_raise  The following rule class moves
  	  tokens between different levels
@@ -706,8 +975,8 @@ r_syll::apply(unit *root)
 class r_raise: public rule
 {
 	virtual OPCODE code() {return OP_RAISE;};
-	bool *whattab;
-	bool *whentab;
+	charclass *whattab;
+	charclass *whentab;
    public:
 		r_raise(char *param);
 	virtual ~r_raise();
@@ -718,14 +987,14 @@ r_raise::r_raise(char *param) : rule(param)
 {
 	char *cond;
 	if ((cond = strchr(raw,RAISE_DELIM))) *cond++=0; else cond=(char *)"!";
-	whattab = booltab(raw);
-	whentab = booltab(cond);
+	whattab = new charclass(raw);
+	whentab = new charclass(cond);
 }
 
 r_raise::~r_raise()
 {
-	free(whattab);
-	free(whentab);
+	delete whattab;
+	delete whentab;
 }
 
 void
@@ -759,13 +1028,13 @@ class r_regex: public rule
 #define PAREN_OPEN  '('
 #define PAREN_CLOS  ')'
 
-#define rshr(x) shriek(811, fmt("Regex invalid: %s%s", x, debug_tag()));
+#define rshr(x) shriek(811, "%s Regex invalid: %s", debug_tag(), x);
 
 r_regex::r_regex(char *param) : rule(param)
 {
 	char separator = *param++;
 	char *tmp = strchr(param, separator);
-	if (!tmp) shriek(811, fmt("regex param should be separated thus: /regex/replacement/ %s", debug_tag()));
+	if (!tmp) shriek(811, "%s Regex param should be separated thus: /regex/replacement/", debug_tag());
 	*tmp++=0;
 	parens = 0;
 	int result;
@@ -785,7 +1054,7 @@ r_regex::r_regex(char *param) : rule(param)
 	param = strdup(scratch);
 
 	matchbuff = (regmatch_t *)xmalloc((parens+2)*sizeof(regmatch_t));
-	DEBUG(0,1,fprintf(STDDBG, "Compiling regex %s\n", param);)
+	D_PRINT(0, "Compiling regex %s\n", param);
 	result = regcomp(&regex, param, 0);
 	switch (result) {
 		case 0: break;
@@ -810,16 +1079,16 @@ r_regex::r_regex(char *param) : rule(param)
 			rshr("too ugly");
 		case REG_ESPACE:
 			rshr("too huge");
-		default: shriek(811, fmt("Bad regex%s, regcomp returns %d", debug_tag(), result));
+		default: shriek(811, "%s Bad regex, regcomp returns %d", debug_tag(), result);
 	}
 	free(param);
 	repl = tmp;
 	tmp = strchr(tmp, separator);
-	if (!tmp) shriek(811, fmt("regex param should be separated thus: /regex/replacement/ %s", debug_tag()));
-	if (tmp[1]) shriek(811, fmt("garbage follows replacement%s", debug_tag()));
+	if (!tmp) shriek(811, "%s Regex param should be separated thus: /regex/replacement/ ", debug_tag());
+	if (tmp[1]) shriek(811, "%s garbage follows replacement", debug_tag());
 	*tmp=0;
 	repl=strdup(repl);
-	DEBUG(0,1,fprintf(STDDBG,"regex%s is okay\n", debug_tag());)
+	D_PRINT(0, "%s Regex is okay\n", debug_tag());
 }
 
 #undef rshr
@@ -846,10 +1115,14 @@ r_regex::apply(unit *root)
  **	  is applied
  ************************************************/
 
+//typedef UINT (CALLBACK* TSR_EVAL)(unit*);
+
+typedef int (*TSR_EVAL)(unit*);
 
 class r_debug: public rule
 {
 	virtual OPCODE code() {return OP_DEBUG;};
+	TSR_EVAL tsr_eval;
    public:
 		r_debug(char *param);
 	virtual void apply(unit *root);
@@ -862,11 +1135,25 @@ r_debug::r_debug(char *param) : rule(param)
 void
 r_debug::apply(unit *root)
 {
+	if(strstr(raw,"tsrtool")) {
+#ifdef HAVE_WINDOWS_H		// The TSR debuging tool is up to now available only in the Win32 port
+		HINSTANCE hDLL = LoadLibrary("tsrtool.dll");
+		if (hDLL != NULL)
+		{
+			tsr_eval = (TSR_EVAL)GetProcAddress(hDLL,"_tsr_eval");
+			if (!tsr_eval) shriek(445, "cannot get 'tsr_eval' function from tsrtool.dll");
+			//root->fout(NULL);
+			//printf("tsr_eval...\n");
+			tsr_eval(root);
+		}
+#endif
+	}
 	if(strstr(raw,"elem")) root->fout(NULL);
 //	if(strstr(raw,"rules")) ruleset->debug();
 //	else if(strstr(raw,"rule") && ruleset->current_rule+1 < ruleset->n_rules)
 //		ruleset->rulist[ruleset->current_rule+1]->debug();
 	if(strstr(raw,"pause")) user_pause();
+//	if(strstr(raw,"dumpunit")) root->filedump ((char*) cfg->dumpfilename);
 }
 
 /************************************************
@@ -911,7 +1198,7 @@ cond_rule::check_children()
 class r_inside: public cond_rule
 {
 	virtual OPCODE code() {return OP_INSIDE;};
-	bool *affected;
+	charclass *affected;
    public:
 		r_inside(char *param, text *file, hash *vars);
 		~r_inside();
@@ -920,18 +1207,63 @@ class r_inside: public cond_rule
 
 r_inside::r_inside(char *param, text *file, hash *vars) : cond_rule(param, file, vars)
 {
-	affected = booltab(raw);
+	affected = new charclass(raw);
 }
 
 r_inside::~r_inside()
 {
-	free(affected);
+	delete affected;
 }
 
 void
 r_inside::apply(unit *root)
 {
-	if (affected[(unsigned char)root->cont]) then->apply(root);
+	if (affected->ismember((unsigned char)root->cont)) then->cook(root);
+}
+
+/************************************************
+ r_near  The following rule class will apply
+ 	  its subordinated rule (block of rules)
+ **	  inside the units which contain a unit
+ **	  whose content is listed in r_near parameter
+ ************************************************/
+
+
+class r_near: public cond_rule
+{
+	virtual OPCODE code() {return OP_NEAR;};
+	bool universal;
+	charclass *affected;
+   public:
+		r_near(char *param, text *file, hash *vars);
+		~r_near();
+	virtual void apply(unit *root);
+};
+
+#define ASTERISK   '*'
+#define ASTERISK_REPLACEMENT	'!'
+
+r_near::r_near(char *param, text *file, hash *vars) : cond_rule(param, file, vars)
+{
+	universal = false;
+	if (*raw == ASTERISK)
+	{
+		universal = true;
+		*raw = ASTERISK_REPLACEMENT;
+	}
+	affected = new charclass(raw);
+	if (universal) *raw = ASTERISK;
+}
+
+r_near::~r_near()
+{
+	delete affected;
+}
+
+void
+r_near::apply(unit *root)
+{
+	if (root->contains(target, affected) ^ universal) then->cook(root);
 }
 
 /************************************************
@@ -946,6 +1278,7 @@ class r_with: public cond_rule
 {
 	virtual OPCODE code() {return OP_WITH;};
 	hash *dict;
+	bool negated;
    public:
 		r_with(char *param, text *file, hash *vars);
 		~r_with();
@@ -956,8 +1289,12 @@ r_with::r_with(char *param, text *file, hash *vars) : cond_rule(param, file, var
 {
 	char *old = raw;
 
-	if (*raw == DQUOT) raw = strdup(raw);
-	else raw = compose_pathname(raw, this_lang->hash_dir, cfg->lang_base_dir);
+	negated = false;
+	if (*raw == '!') raw++, negated = true;
+
+	raw = strdup(raw);
+//	if (*raw == DQUOT) raw = strdup(raw);
+//	else raw = compose_pathname(raw, this_lang->hash_dir, scfg->lang_base_dir);
 
 	free(old);
 	dict = NULL;
@@ -973,12 +1310,12 @@ r_with::apply(unit *root)
 {
 	if (!dict) {
 		if (*raw == DQUOT) dict = literal_hash(raw);
-		else dict = new hash(raw, cfg->hash_full,
-			0, 200, 3, DATA_EQUALS_KEY, false, "dictionary %s not found", esctab);
+		else dict = new hash(raw, this_lang->hash_dir, scfg->lang_base_dir, "dictionary",
+			scfg->hashes_full, 0, 200, 3, DATA_EQUALS_KEY, false);
 	}
-	if (!dict) shriek(811, fmt("Unterminated argument%s", debug_tag()));	// or out of memory
+	if (!dict) shriek(811, "%s Unterminated argument", debug_tag());	// or out of memory
 
-	if (root->subst(dict, M_ONCE)) then->apply(root);
+	if (root->subst(dict, NULL, M_ONCE) ^ negated) then->cook(root);
 }
 
 
@@ -1002,10 +1339,10 @@ class r_if: public cond_rule
 
 r_if::r_if(char *param, text *file, hash *vars) : cond_rule(param, file, vars)
 {
-	option *o = option_struct(raw + (*raw == EXCLAM) , this_lang->soft_options);
-	if (!o) shriek(811, fmt("Not an option: %s%s", raw, debug_tag()));
-	if (o->opttype != O_BOOL) shriek(811, fmt("Not a truth value option: %s%s", raw, debug_tag()));
-	if (o->structype != OS_VOICE) shriek(811, fmt("Not a voice option: %s%s", raw, debug_tag()));
+	epos_option *o = option_struct(raw + (*raw == EXCLAM) , this_lang->soft_opts);
+	if (!o) shriek(811, "%s Not an option: %s", debug_tag(), raw);
+	if (o->opttype != O_BOOL) shriek(811, "%s Not a truth value option: %s", debug_tag(), raw);
+	if (o->structype != OS_VOICE) shriek(811, "%s Not a voice option: %s", debug_tag(), raw);
 	flag_offs = o->offset;
 }
 
@@ -1013,7 +1350,7 @@ void
 r_if::apply(unit *root)		// this_lang must correspond to these rules!
 {
 	if (this_voice && (*(bool *)((char *)this_voice + flag_offs)) ^ (*raw == EXCLAM))
-		then->apply(root);
+		then->cook(root);
 }
 
 void
@@ -1040,5 +1377,46 @@ class r_nothing: public rule
 	virtual void apply(unit *) {};
 };
 
+
+/************************************************
+ r_neural  Neuralnet created by Jakub Adamek
+ 	    look at neural.cc, neural.h, unit.cc
+ **	 
+ ************************************************/
+
+#include "nnet/neural.h"
+
+class r_neural: public rule
+{
+protected:
+	virtual OPCODE code() {return OP_NNET;};
+	CNeuralNet *neuralnet;
+public:
+	r_neural (char *param, hash *vars);
+	virtual ~r_neural ();
+	virtual void apply(unit *root);
+};
+
+void
+r_neural::apply (unit *root)
+{
+	if (!neuralnet)
+		shriek(861, "The --neuronet option is false");
+	neuralnet->init();
+	root->neural (target, neuralnet);
+}
+
+r_neural::r_neural(char *param, hash *vars) : rule(param)
+{
+	neuralnet = NULL;
+	if (scfg->neuronet)
+		neuralnet = new CNeuralNet(raw, vars);
+}
+
+r_neural::~r_neural ()
+{
+	if (neuralnet) delete neuralnet;
+}
+
 #include "block.cc"
 
