@@ -1,7 +1,7 @@
 /*
  *	epos/src/waveform.cc
- *	(c) 1998-01 geo@cuni.cz
- *	(c) 2001 horak@ure.cas.cz
+ *	(c) 1998-02 geo@cuni.cz
+ *	(c) 2000-02 horak@ure.cas.cz
  *
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -15,7 +15,7 @@
  *
  */
 
-#include "common.h"
+#include "epos.h"
 #include "client.h"
 
 #ifdef HAVE_FCNTL_H
@@ -85,8 +85,10 @@ static int downsample_factor(int working, int out)
 	if (!out) return 1;
 	if (working == out) return 1;
 	if (working / 2 == out) return 2;
-	shriek(462, fmt("Currently, you can only get %d Hz and %d Hz output signal with this voice",
-		working, working / 2));
+	if (working / 3 == out) return 3;
+	if (working / 4 == out) return 4;
+	shriek(462, "Currently, you can only get %d Hz, %d Hz, %d Hz or %d Hz output signal with this voice",
+		working, working / 2, working / 3, working / 4);
 	return 1;
 }
 
@@ -96,17 +98,17 @@ static int downsample_factor(int working, int out)
 //#pragma hdrstop
 
 // #define RIFF_HEADER_SIZE  8
-#define WAVE_HEADER_SIZE  ((long)(sizeof(wave_header) - RIFF_HEADER_SIZE))
+#define WAVE_HEADER_SIZE  ((int32_t)(sizeof(wave_header) - RIFF_HEADER_SIZE))
 
 
 struct cue_point
 {
-	long name;
-	long pos;
+	int32_t name;
+	int32_t pos;
 	char chunk[4];
-	long chunkstart;
-	long blkstart;
-	long sample_offset;
+	int32_t chunkstart;
+	int32_t blkstart;
+	int32_t sample_offset;
 };
 
 cue_point cue_point_template = { 0, 0, FOURCC_INIT("data"), 0, 0, 0 };
@@ -117,44 +119,46 @@ cue_point cue_point_template = { 0, 0, FOURCC_INIT("data"), 0, 0, 0 };
 struct ltxt
 {
 	char txt[4];
-	long len;
-	long cp_name;
-	long sample_count;
+	int32_t len;
+	int32_t cp_name;
+	int32_t sample_count;
 	char purpose[4];
-	short country;
-	short language;
-	short dialect;
-	short codepage;
+	int16_t country;
+	int16_t language;
+	int16_t dialect;
+	int16_t codepage;
 };
 
 struct labl
 {
 	char txt[4];
-	long len;
-	long cp_name;
+	int32_t len;
+	int32_t cp_name;
 };
 
-#define USA	 1		// FIXME (etc.)
-#define English  9
-#define American 1
-#define Boring_CodePage	437
 
-ltxt ltxt_template = { FOURCC_INIT("ltxt"), 0, 0, 0, FOURCC_INIT("dphl"), USA, English, American, Boring_CodePage };
+
+//#define USA	 1		// FIXME (etc.)
+//#define English  9
+//#define American 1
+//#define Boring_CodePage	437
+
+//ltxt ltxt_template = { FOURCC_INIT("ltxt"), 0, 0, 0, FOURCC_INIT("dphl"), USA, English, American, Boring_CodePage };
 labl labl_template = { FOURCC_INIT("labl"), 0, 0 };
 labl note_template = { FOURCC_INIT("note"), 0, 0 };
 
 wavefm::wavefm(voice *v)
 {
-	samp_rate = v->samp_rate;
+	samp_rate = v->inv_sampling_rate;
 //	samp_size_bytes = sizeof(SAMPLE);
 	channel = v->channel;
 	int stereo = channel==CT_MONO ? 0 : 1;
-	if (this_voice->samp_size <= 0 || (this_voice->samp_size >> 3) > (signed)sizeof(int))
-		shriek(447, fmt("Invalid sample size %d", this_voice->samp_size));
-	if (cfg->ulaw && !cfg->wav_hdr)
-		this_voice->samp_size = 8, this_voice->out_rate = 8000;
-//	cfg->samp_size = cfg->samp_size + 7 & ~7;	// Petr had this
-	downsamp = downsample_factor(samp_rate, v->out_rate);
+	if (this_voice->sample_size <= 0 || (this_voice->sample_size >> 3) > (signed)sizeof(int))
+		shriek(447, "Invalid sample size %d", this_voice->sample_size);
+	if (cfg->ulaw && !cfg->wave_header)
+		this_voice->sample_size = 8, this_voice->out_sampling_rate = 8000;
+//	cfg->sample_size = cfg->sample_size + 7 & ~7;	// Petr had this
+	downsamp = downsample_factor(samp_rate, v->out_sampling_rate);
 	translated = false;
 
 	memcpy(hdr.string1, "RIFF", 4);
@@ -164,8 +168,13 @@ wavefm::wavefm(voice *v)
 	hdr.numchan = stereo ? 2 : 1;
 	hdr.sf1 = samp_rate; hdr.sf2 = stereo ? hdr.sf1 : 0;
 	hdr.avr1 = 2 * samp_rate; hdr.avr2 = stereo ? hdr.avr1 : 0;
+
+//	hdr.wlenB = samp_size_bytes; hdr.wlenb = v->sample_size;  <--pre 2.4.66 minus 2.4.40
+//	hdr.xnone = 0x010;					<--dtto
+
 	hdr.alignment = sizeof(SAMPLE); hdr.samplesize = sizeof(SAMPLE) << 3;
 	hdr.fmt_length = 0x010;
+
 //	written_bytes = 0;
 	hdr.total_length = - RIFF_HEADER_SIZE;
 //	hdr.total_length = hdr.written_bytes - RIFF_HEADER_SIZE;
@@ -180,8 +189,8 @@ wavefm::wavefm(voice *v)
 	buffer = NULL;
 	hdr.buffer_idx = 0;
 
-	cuehdr.len = 4;
 	cuehdr.n = adtlhdr.len = 0;
+	cuehdr.len = 4;
 	memcpy(cuehdr.string1, "cue ", 4);
 	memcpy(adtlhdr.string1, "LIST", 4);
 	memcpy(adtlhdr.string2, "adtl", 4);
@@ -253,7 +262,7 @@ static inline void set_samp_size(int fd, int samp_size_bits)
    #ifdef SNDCTL_DSP_GETFMTS
 	int mask = (unsigned int)-1;
 	ioctl (fd, SNDCTL_DSP_GETFMTS, &mask);
-	DEBUG(3,9,fprintf(STDDBG,"Hardware format mask is 0x%04x\n", mask);)
+	D_PRINT(3, "Hardware format mask is 0x%04x\n", mask);
 	if (!(samp_size_bits & mask)) shriek(439, "Sampling rate not supported");
    #endif
 }
@@ -316,7 +325,7 @@ static const inline bool ioctlable(int fd)
 
 static const inline bool ioctlable(int fd)
 {
-	DEBUG(2,9,fprintf(STDDBG, "Sound ioctl's absent\n");)
+	D_PRINT(2, "Sound ioctl's absent\n");
 	return false;
 }
 
@@ -407,9 +416,9 @@ void
 wavefm::detach(int)
 {
 	while (flush()) ;		// FIXME: think slow network write
-	DEBUG(2,9,fprintf(STDDBG,"Detaching waveform %s\n", ""););
+	D_PRINT(2, "Detaching waveform\n");
 	sync_soundcard(fd);		// FIXME: necessary, but unwanted
-	if (cfg->wav_hdr && !ioctlable(fd)) write_header();
+	if (cfg->wave_header && !ioctlable(fd)) write_header();
 	fd = -1;
 }
 
@@ -424,16 +433,16 @@ wavefm::brk()
 void
 wavefm::attach(int d)
 {
-	DEBUG(2,9,fprintf(STDDBG,"Attaching waveform %s\n", ""););
+	D_PRINT(2, "Attaching waveform\n");
 	fd = d;
 	translate();
 	hdr.total_length = get_total_len() - RIFF_HEADER_SIZE;
 	if (ioctlable(fd)) {
 		ioctl_attach();
 	}
-	DEBUG(0,9,fprintf(STDDBG,"(attached, now flushing predata)\n");)
+	D_PRINT(0, "(attached, now flushing predata)\n");
 	if (hdr.buffer_idx) flush();
-	DEBUG(0,9,fprintf(STDDBG,"(predata flushed)\n");)
+	D_PRINT(0, "(predata flushed)\n");
 	if (buff_size) buffer = buffer ? (SAMPLE *)xrealloc(buffer, buff_size * sizeof(SAMPLE)) : (SAMPLE *)xmalloc(buff_size);
 }
 
@@ -444,8 +453,8 @@ wavefm::attach()
 	int d;
 
 	if (fd != -1) shriek(862, "Nested voice::attach()");
-	if (!cfg->play_segs) cfg->local_sound_device = NULL_FILE;
-	output = compose_pathname(cfg->local_sound_device, cfg->wav_dir);
+	if (!scfg->play_segments) scfg->local_sound_device = NULL_FILE;
+	output = compose_pathname(scfg->local_sound_device, scfg->wav_dir);
 
 #ifdef S_IRGRP
 	d = open(output, O_WRONLY | O_CREAT | O_TRUNC | O_NONBLOCK | O_BINARY, MODE_MASK);
@@ -455,8 +464,8 @@ wavefm::attach()
 #ifdef	HAVE_MMSYSTEM_H
 	localsound = d;
 #endif
-	if (d == -1) shriek(445, fmt("Failed to %s %s", strncmp(output, "/dev/", 5)
-			? "create output file" : "open audio device", output));
+	if (d == -1) shriek(445, "Failed to %s %s", strncmp(output, "/dev/", 5)
+			? "create output file" : "open audio device", output);
 	free(output);
 	attach(d);
 }
@@ -483,44 +492,55 @@ static const int exp_lut[256] = {0,0,1,1,2,2,2,2,3,3,3,3,3,3,3,3,4,4,4,4,4,4,4,4
 /* The following is a fixed (FIXME: generalize) low-pass band filter, 0 to 4 kHz */
 
 void
-wavefm::band_filter()
+wavefm::band_filter(int ds)
 {
-	const double a[9] = {1,-2.95588833875285,6.05358859038796,-8.21891083745587,8.44331226032319,-6.3984876438294,3.5670260078076,-1.33913357618129,0.282464627918022};
-	const double b[9] = {0.0085475284871725,0.0230290518801368,0.0495486106178739,0.0712620928385417,0.0820025736317015,0.0712620928385417,0.049548610617874,0.0230290518801368,0.00854752848717252};
+	const double a[3][9] = {{1,-2.95588833875285,6.05358859038796,
+				 -8.21891083745587,8.44331226032319,-6.3984876438294,
+				 3.5670260078076,-1.33913357618129,0.282464627918022},
+				{1,-5.64288854258764,15.23495961198290,
+				 -25.26198227870461,27.94306848298673,-21.03831667925559,
+				 10.52011230037758,-3.19916382894058,0.45524725438187},
+				{1,-6.42466721653009,18.92807679520927,
+				 -33.22220198267817,37.88644682915341,-28.70015825434698,
+				 14.09511202954144,-4.10429644762400,0.54321917883107}};
+	const double b[3][9] = {{0.0085475284871725,0.0230290518801368,0.0495486106178739,
+				0.0712620928385417,0.0820025736317015,0.0712620928385417,
+				0.049548610617874,0.0230290518801368,0.00854752848717252},
+				{0.00182487646825,-0.00196817471641,0.00491609616544,
+				-0.00250076958374,0.00529207410096,-0.00250076958374,
+				0.00491609616544,-0.00196817471641,0.00182487646825},
+				{0.00101813708883,-0.00274510939089,0.00509342597553,
+				 -0.00601560729994,0.00666275143843,-0.00601560729994,
+				 0.00509342597553,-0.00274510939089,0.00101813708883}};
 	
 	static double filt[9];
 	int i,j;
 	
-	DEBUG(1,9,fprintf(STDDBG,"Low-pass filter is applied\n");)
+	D_PRINT(1, "Low-pass filter is applied\n");
 	for (i = 0; i < 9; i++) filt[i] = 0;
 
 	for (i = 0; i < hdr.buffer_idx; i++) {
-		filt[0] = (double)(SIGNED_SAMPLE)buffer[i];
-		for (j=1;j<9;filt[0]=filt[0]-a[j]*filt[j++]);
+		filt[0] = (double)buffer[i];
+		for (j=1;j<9;filt[0]=filt[0]-a[ds-2][j]*filt[j++]);
 		double outsamp = 0;
-		for (j=0;j<9;outsamp=outsamp+b[j]*filt[j++]);
-//		printf("%d %f        ", buffer[i], (float)outsamp);
-		buffer[i] = (SAMPLE)(SIGNED_SAMPLE)outsamp;
+		for (j=0;j<9;outsamp=outsamp+b[ds-2][j]*filt[j++]);
+		D_PRINT(0, "%d %f        ", buffer[i], (float)outsamp);
+		buffer[i] = (SAMPLE)outsamp;
 		for (j=8;j>0;j--) filt[j]=filt[j-1];		//FIXME: speed up
 	}
+	D_PRINT(0, "\n");
 }
 
-
-
-// #define put_sample_8(sample)  *newbuff++ = (char)sample
-// #define put_sample_16(sample) *(unsigned short *)newbuff = sample, newbuff += 2
-// #define put_sample(sample) if (eight_bit) put_sample_8(sample); else put_sample_16(sample)
-
 #define put_sample(sample) *(int *)newbuff = sample, newbuff += ssbytes;
 
 inline void
 wavefm::translate_data(char *newbuff)
 {
-	DEBUG(1,9,fprintf(STDDBG,"(translating the data, too)");)
-//	if (cfg->ulaw && this_voice->samp_size != 8) shriek(462, "Mu law implies 8 bit");
-	int ssbytes  = (this_voice->samp_size + 7) >> 3;  // sample size in bytes
+	D_PRINT(1, "(translating the data, too)");
+//	if (cfg->ulaw && this_voice->sample_size != 8) shriek(462, "Mu law implies 8 bit");
+	int ssbytes  = (this_voice->sample_size + 7) >> 3;  // sample size in bytes
 	int shift1 = (sizeof(int) - sizeof(SAMPLE)) << 3;
-	int shift2 = cfg->big_endian ? 0 : (sizeof(int) - ssbytes) << 3;
+	int shift2 = scfg->_big_endian ? 0 : (sizeof(int) - ssbytes) << 3;
 	int unsign = ssbytes == 1 ? 0x80 : 0;
 	for (int i = 0; i < hdr.buffer_idx; i += downsamp) {
 		int sample = buffer[i];
@@ -548,16 +568,21 @@ wavefm::translate_data(char *newbuff)
 void
 wavefm::translate()
 {
-	DEBUG(1,9,fprintf(STDDBG,"Translating waveform, buffer_idx=%d\n", hdr.buffer_idx);)
+	D_PRINT(1, "Translating waveform, buffer_idx=%d\n", hdr.buffer_idx);
 	if (translated || !hdr.buffer_idx) return;
 	
 	int working_size = sizeof (SAMPLE);
 	working_size *= downsamp;
 
-	int target_size = this_voice->samp_size >> 3;
+	int target_size = this_voice->sample_size >> 3;
 	target_size *= (1 + (channel != CT_MONO));
 	
-	if (downsamp != 1 && cfg->autofilter) band_filter();	//output is downsampled
+//	if (downsamp != 1 && cfg->autofilter) band_filter();	//output is downsampled
+	if (cfg->autofilter) {
+		if (downsamp == 2) band_filter(downsamp);	//output is downsampled
+		if (downsamp == 3) band_filter(downsamp);	//output is downsampled
+		if (downsamp == 4) band_filter(downsamp);	//output is downsampled
+	}
 	
 	if (downsamp == 1 && working_size == target_size && channel == CT_MONO
 						&& !cfg->ulaw) goto finis;
@@ -570,14 +595,14 @@ wavefm::translate()
 		translate_data((char *)buffer);		// strange semantics
 	}
 	
-	if (this_voice->out_rate) samp_rate = this_voice->out_rate;
+	if (this_voice->out_sampling_rate) samp_rate = this_voice->out_sampling_rate;
 	hdr.sf1 = samp_rate;		if (hdr.sf2) hdr.sf2 = hdr.sf1;
 	hdr.avr1 = samp_rate * target_size;	if (hdr.avr2) hdr.avr2 = hdr.avr1;
-	hdr.alignment = target_size;	hdr.samplesize = this_voice->samp_size;
+	hdr.alignment = target_size;	hdr.samplesize = this_voice->sample_size;
 	if (cfg->ulaw) hdr.datform = IBM_FORMAT_MULAW;
    finis:
 	translated = true;
-	hdr.buffer_idx = hdr.buffer_idx * target_size / downsamp;
+	hdr.buffer_idx = (hdr.buffer_idx + 1) / downsamp * target_size;
 }
 
 /*
@@ -600,7 +625,7 @@ wavefm::translate()
 struct w_ophase
 {
 	bool inlined;
-	int  adjustment;
+	int adjustment;
 	char **buff;
 	int  *len;
 };
@@ -616,7 +641,7 @@ const w_ophase wavefm::ophases[] = {
 	INLINED_WOPH(cuehdr, cue_header)
 	VAR_WOPH(cp_buff, cuehdr.len, -4)
 	INLINED_WOPH(adtlhdr, adtl_header)
-	VAR_WOPH(adtl_buff, adtlhdr.len, 0)
+	VAR_WOPH(adtl_buff, adtlhdr.len, 0)	// last must not be inlined! else skipped
 	{true, 0, WOPHASE_NO_MORE_BUFFS, (int *)0}
 };
 
@@ -651,7 +676,7 @@ wavefm::get_ophase_len(const w_ophase *p)
 inline bool
 wavefm::update_ophase()
 {
-	if (!cfg->wav_hdr || (fd != -1 && ioctlable(fd))) {	// sound card treated specially
+	if (!cfg->wave_header || (fd != -1 && ioctlable(fd))) {	// sound card treated specially
 		if (ophase == 0) {
 			ophase++, ooffset = 0;
 		}
@@ -686,14 +711,15 @@ wavefm::flush()
 	if (fd == -1)
 		return flush_deferred();
 	translate();
-		
-	written = ywrite(fd, get_ophase_buff(&ophases[ophase]) + ooffset,
-			get_ophase_len(&ophases[ophase]) - ooffset);
+
+	const char *o_buff = get_ophase_buff(&ophases[ophase]) + ooffset;
+	int o_buff_size = get_ophase_len(&ophases[ophase]) - ooffset;
+	written = ywrite(fd, o_buff, o_buff_size);
 	if (1 > written) return flush_deferred();
 	
 	ooffset += written;
 
-	DEBUG(2,9,fprintf(STDDBG, "Flushing the signal\n");)
+	D_PRINT(2, "Flushing the signal\n");
 //	hdr.total_length += written;
 	return true;
 }
@@ -701,8 +727,8 @@ wavefm::flush()
 bool
 wavefm::flush_deferred()
 {
-	DEBUG(2,9,fprintf(STDDBG, "Flushing the signal (deferred)\n");)
-	DEBUG(1,9,fprintf(STDDBG, fmt("adtlhdr.len is %d\n", adtlhdr.len));)
+	D_PRINT(2, "Flushing the signal (deferred)\n");
+	D_PRINT(1, "adtlhdr.len is %d\n", adtlhdr.len);
 	written = 0;
 	if (hdr.buffer_idx + 4 <= buff_size)
 		return true;
@@ -740,7 +766,8 @@ wavefm::put_chunk(labl *chunk_template, const char *string)
 	*la = *chunk_template;
 	la->len = sizeof(labl) + varlen - RIFF_HEADER_SIZE;
 	la->cp_name = current_cp;
-	strcpy((char *)(la+1), string);
+	strcpy((char *)(la + 1), string);
+	decode_string((char *)(la + 1), this_lang->charset);
 	((char *)(la+1))[varlen - 1] = 0;	// padding if odd
 	adtlhdr.len += sizeof(labl) + varlen;
 }
@@ -756,12 +783,20 @@ wavefm::label(int pos, char *label, const char *note)
 	cp_buff[current_cp] = cue_point_template;
 	cp_buff[current_cp].name = current_cp + 1;	/* numbered starting from 1 */
 	cp_buff[current_cp].pos = cp_buff[current_cp].sample_offset = hdr.buffer_idx - pos;
+
+
+// FIXME! choose one of the following three solutions
+
+//	cp_buff[current_cp].pos /= samp_size_bytes;	// pos in samples, offset in bytes
+//	cp_buff[current_cp].pos /= sizeof(SAMPLE);	// pos in samples, offset in bytes
+//	cp_buff[current_cp].pos /= 1;			// pos in samples, offset in bytes
+
 	current_cp++;
 	cuehdr.len += sizeof(cue_point);
 	cuehdr.n++;
-
+	
 	if (!label) return;
-
+	
 	put_chunk(&labl_template, label);
 	put_chunk(&note_template, note);
 }
@@ -772,14 +807,14 @@ void
 wavefm::become(void *b, int size)
 {
 	hdr = *(wave_header *)b;
+	hdr.total_length = size - HEADER_HEADER_SIZE;
 	size -= sizeof(wave_header);
-	if (size < 0) shriek(471, "tcpsyn got garbage");
+	if (size < 0) shriek(471, "tcpsyn or mbrsyn got garbage");
 	buffer = (char *)xmalloc(size);
 	memcpy(buffer, (char *)b + sizeof(wave_header), size);
 	buff_size = size;
 	hdr.buffer_idx = size;
 //	written_bytes = 0;
-	hdr.total_length = 0 - HEADER_HEADER_SIZE;
 	channel = hdr.sf2 ? CT_BOTH : CT_MONO;
 	samp_rate = hdr.sf1;
 }
@@ -790,27 +825,45 @@ wavefm::become(void *b, int size)
 
 #define CHUNK_HEADER_SIZE 8
 
-void pchu(char *p, int size)
+void
+wavefm::chunk_become(char *p, int size)
 {
 	int l = *((int *)p+1);
-//	printf("pchu %p+%d, %.4s len %d\n", p, size, p, *((int *)p+1));
+	D_PRINT(2, "Considering chunk %p+%d, %.4s len %d\n", p, size, p, *((int *)p+1));
 	if (FOURCC_ID(p) == FOURCC_ID("RIFF")) {
-//		printf("RIFF here\n");
+		D_PRINT(0, "RIFF here\n");
 		if (FOURCC_ID(p + RIFF_HEADER_SIZE) != FOURCC_ID("WAVE")) shriek(471, "Other RIFF than WAVE received");
+		l -= 4;
 		for (char *q = p+12; l>0; ) {
 			int chunksize = ((int *)q)[1];
-			pchu(q, chunksize);
+			chunk_become(q, chunksize);
 			l -= chunksize + CHUNK_HEADER_SIZE;
 			q += chunksize + CHUNK_HEADER_SIZE;
 		}
+		return;
 	}
+	if (FOURCC_ID(p) == FOURCC_ID("fmt ")) {
+		hdr = *(wave_header *)(p - 12);
+		channel = hdr.sf2 ? CT_BOTH : CT_MONO;
+		samp_rate = hdr.sf1;
+		return;
+	}
+	if (FOURCC_ID(p) == FOURCC_ID("data")) {
+		buffer = (SAMPLE *)xmalloc(size);
+		memcpy(buffer, p + 8, size);
+		buff_size = size;
+		hdr.buffer_idx = size;
+		translated = true;
+		return;
+	}
+	shriek(471, "Unknown chunk %.4s", p);
 }
 
 void wavefm::become(void *b, int size)
 {
 	((wave_header *)b)->total_length = size - RIFF_HEADER_SIZE;
 	((wave_header *)b)->buffer_idx = size - sizeof(wave_header);
-	pchu((char *)b, size);
+	chunk_become((char *)b, size);
 }
 
 #endif
